import pathlib
import subprocess
import sys

import pytest

from vulture import core

REPO = pathlib.Path(__file__).resolve().parents[1]
WHITELISTS = [
    str(path) for path in (REPO / "vulture" / "whitelists").glob("*.py")
]


def call_vulture(args, **kwargs):
    return subprocess.call(
        [sys.executable, "-m", "vulture"] + args, cwd=REPO, **kwargs
    )


def check(items_or_names, expected_names):
    """items_or_names must be a collection of Items or a set of strings."""
    try:
        assert sorted(item.name for item in items_or_names) == sorted(
            expected_names
        )
    except AttributeError:
        assert items_or_names == set(expected_names)


def check_unreachable(v, lineno, size, name):
    assert len(v.unreachable_code) == 1
    item = v.unreachable_code[0]
    assert item.first_lineno == lineno
    assert item.size == size
    assert item.name == name


@pytest.fixture
def v():
    return core.Vulture(verbose=True)



import ast
import pathlib
import sys
import tokenize


class VultureInputException(Exception):
    pass


def _safe_eval(node, default):
    """
    Safely evaluate the Boolean expression under the given AST node.

    Substitute `default` for all sub-expressions that cannot be
    evaluated (because variables or functions are undefined).

    We could use eval() to evaluate more sub-expressions. However, this
    function is not safe for arbitrary Python code. Even after
    overwriting the "__builtins__" dictionary, the original dictionary
    can be restored
    (https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html).

    """
    if isinstance(node, ast.BoolOp):
        results = [_safe_eval(value, default) for value in node.values]
        if isinstance(node.op, ast.And):
            return all(results)
        else:
            return any(results)
    elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not):
        return not _safe_eval(node.operand, not default)
    else:
        try:
            return ast.literal_eval(node)
        except ValueError:
            return default


def condition_is_always_false(condition):
    return not _safe_eval(condition, True)


def condition_is_always_true(condition):
    return _safe_eval(condition, False)


def format_path(path):
    try:
        return path.relative_to(pathlib.Path.cwd())
    except ValueError:
        # Path is not below the current directory.
        return path


def get_decorator_name(decorator):
    if isinstance(decorator, ast.Call):
        decorator = decorator.func
    parts = []
    while isinstance(decorator, ast.Attribute):
        parts.append(decorator.attr)
        decorator = decorator.value
    parts.append(decorator.id)
    return "@" + ".".join(reversed(parts))


def get_modules(paths):
    """Retrieve Python files to check.

    Loop over all given paths, abort if any ends with .pyc, add the other given
    files (even those not ending with .py) and collect all .py files under the
    given directories.

    """
    modules = []
    for path in paths:
        path = path.resolve()
        if path.is_file():
            if path.suffix == ".pyc":
                sys.exit(f"Error: *.pyc files are not supported: {path}")
            else:
                modules.append(path)
        elif path.is_dir():
            modules.extend(path.rglob("*.py"))
        else:
            sys.exit(f"Error: {path} could not be found.")
    return modules


def read_file(filename):
    try:
        # Use encoding detected by tokenize.detect_encoding().
        with tokenize.open(filename) as f:
            return f.read()
    except (SyntaxError, UnicodeDecodeError) as err:
        raise VultureInputException(err)


class LoggingList(list):
    def __init__(self, typ, verbose):
        self.typ = typ
        self._verbose = verbose
        return super().__init__()

    def append(self, item):
        if self._verbose:
            print(f'define {self.typ} "{item.name}"')
        super().append(item)


class LoggingSet(set):
    def __init__(self, typ, verbose):
        self.typ = typ
        self._verbose = verbose
        return super().__init__()

    def add(self, name):
        if self._verbose:
            print(f'use {self.typ} "{name}"')
        super().add(name)



import codecs

from . import v

assert v  # Silence pyflakes.


def test_encoding1(v):
    v.scan(
        """\
# -*- coding: utf-8 -*-
pass
"""
    )
    assert not v.found_dead_code_or_error


def test_encoding2(v):
    v.scan(
        """\
#! /usr/bin/env python
# -*- coding: utf-8 -*-
pass
"""
    )
    assert not v.found_dead_code_or_error


def test_non_utf8_encoding(v, tmp_path):
    code = ""
    name = "non_utf8"
    non_utf_8_file = tmp_path / (name + ".py")
    with open(non_utf_8_file, mode="wb") as f:
        f.write(codecs.BOM_UTF16_LE)
        f.write(code.encode("utf_16_le"))
    v.scavenge([non_utf_8_file])
    assert v.found_dead_code_or_error


def test_utf8_with_bom(v, tmp_path):
    name = "utf8_bom"
    filepath = tmp_path / (name + ".py")
    # utf8_sig prepends the BOM to the file.
    filepath.write_text("", encoding="utf-8-sig")
    v.scavenge([filepath])
    assert not v.found_dead_code_or_error



import pytest

from . import v

assert v  # Silence pyflakes

mock_code = """\
import foo

class Foo:
    def __init__(self):
        print("Initialized foo")

    def bar(self):
        self.foobar = "unused attribute"
        foobar = "unused variable"
        return
        print("unreachable")

    @property
    def myprop(self):
        pass

def myfunc():
    pass
"""


@pytest.fixture
def check_report(v, capsys):
    def test_report(code, expected, make_whitelist=False):
        filename = "foo.py"
        v.scan(code, filename=filename)
        capsys.readouterr()
        v.report(make_whitelist=make_whitelist)
        assert capsys.readouterr().out == expected.format(filename=filename)

    return test_report


def test_item_report(check_report):
    expected = """\
{filename}:1: unused import 'foo' (90% confidence)
{filename}:3: unused class 'Foo' (60% confidence)
{filename}:7: unused method 'bar' (60% confidence)
{filename}:8: unused attribute 'foobar' (60% confidence)
{filename}:9: unused variable 'foobar' (60% confidence)
{filename}:11: unreachable code after 'return' (100% confidence)
{filename}:13: unused property 'myprop' (60% confidence)
{filename}:17: unused function 'myfunc' (60% confidence)
"""
    check_report(mock_code, expected)


def test_make_whitelist(check_report):
    expected = """\
foo  # unused import ({filename}:1)
Foo  # unused class ({filename}:3)
_.bar  # unused method ({filename}:7)
_.foobar  # unused attribute ({filename}:8)
foobar  # unused variable ({filename}:9)
# unreachable code after 'return' ({filename}:11)
_.myprop  # unused property ({filename}:13)
myfunc  # unused function ({filename}:17)
"""
    check_report(mock_code, expected, make_whitelist=True)



from unittest import mock, TestCase

TestCase.setUp
TestCase.tearDown
TestCase.setUpClass
TestCase.tearDownClass
TestCase.run
TestCase.skipTest
TestCase.debug
TestCase.failureException
TestCase.longMessage
TestCase.maxDiff
TestCase.subTest

mock.Mock.return_value
mock.Mock.side_effect



"""
Unit tests for config file and CLI argument parsing.
"""

from io import StringIO
from textwrap import dedent

import pytest

from vulture.config import (
    DEFAULTS,
    _check_input_config,
    _parse_args,
    _parse_toml,
    make_config,
)


def test_cli_args():
    """
    Ensure that CLI arguments are converted to a config object.
    """
    expected = dict(
        paths=["path1", "path2"],
        exclude=["file*.py", "dir/"],
        ignore_decorators=["deco1", "deco2"],
        ignore_names=["name1", "name2"],
        make_whitelist=True,
        min_confidence=10,
        sort_by_size=True,
        verbose=True,
    )
    result = _parse_args(
        [
            "--exclude=file*.py,dir/",
            "--ignore-decorators=deco1,deco2",
            "--ignore-names=name1,name2",
            "--make-whitelist",
            "--min-confidence=10",
            "--sort-by-size",
            "--verbose",
            "path1",
            "path2",
        ]
    )
    assert isinstance(result, dict)
    assert result == expected


def test_toml_config():
    """
    Ensure parsing of TOML files results in a valid config object.
    """
    expected = dict(
        paths=["path1", "path2"],
        exclude=["file*.py", "dir/"],
        ignore_decorators=["deco1", "deco2"],
        ignore_names=["name1", "name2"],
        make_whitelist=True,
        min_confidence=10,
        sort_by_size=True,
        verbose=True,
    )
    data = StringIO(
        dedent(
            """\
        [tool.vulture]
        exclude = ["file*.py", "dir/"]
        ignore_decorators = ["deco1", "deco2"]
        ignore_names = ["name1", "name2"]
        make_whitelist = true
        min_confidence = 10
        sort_by_size = true
        verbose = true
        paths = ["path1", "path2"]
        """
        )
    )
    result = _parse_toml(data)
    assert isinstance(result, dict)
    assert result == expected


def test_config_merging():
    """
    If we have both CLI args and a ``pyproject.toml`` file, the CLI args should
    have precedence.
    """
    toml = StringIO(
        dedent(
            """\
        [tool.vulture]
        exclude = ["toml_exclude"]
        ignore_decorators = ["toml_deco"]
        ignore_names = ["toml_name"]
        make_whitelist = false
        min_confidence = 10
        sort_by_size = false
        verbose = false
        paths = ["toml_path"]
        """
        )
    )
    cliargs = [
        "--exclude=cli_exclude",
        "--ignore-decorators=cli_deco",
        "--ignore-names=cli_name",
        "--make-whitelist",
        "--min-confidence=20",
        "--sort-by-size",
        "--verbose",
        "cli_path",
    ]
    result = make_config(cliargs, toml)
    expected = dict(
        paths=["cli_path"],
        exclude=["cli_exclude"],
        ignore_decorators=["cli_deco"],
        ignore_names=["cli_name"],
        make_whitelist=True,
        min_confidence=20,
        sort_by_size=True,
        verbose=True,
    )
    assert result == expected


def test_config_merging_missing():
    """
    If we have set a boolean value in the TOML file, but not on the CLI, we
    want the TOML value to be taken.
    """
    toml = StringIO(
        dedent(
            """\
        [tool.vulture]
        verbose = true
        ignore_names = ["name1"]
        """
        )
    )
    cliargs = [
        "cli_path",
    ]
    result = make_config(cliargs, toml)
    assert result["verbose"] is True
    assert result["ignore_names"] == ["name1"]


def test_config_merging_toml_paths_only():
    """
    If we have paths in the TOML but not on the CLI, the TOML paths should be
    used.
    """
    toml = StringIO(
        dedent(
            """\
        [tool.vulture]
        paths = ["path1", "path2"]
        """
        )
    )
    cliargs = [
        "--exclude=test_*.py",
    ]
    result = make_config(cliargs, toml)
    assert result["paths"] == ["path1", "path2"]
    assert result["exclude"] == ["test_*.py"]


def test_invalid_config_options_output():
    """
    If the config file contains unknown options we want to abort.
    """

    with pytest.raises(SystemExit):
        _check_input_config({"unknown_key_1": 1})


@pytest.mark.parametrize("key, value", list(DEFAULTS.items()))
def test_incompatible_option_type(key, value):
    """
    If a config value has a different type from the default value we abort.
    """
    wrong_types = {int, str, list, bool} - {type(value)}
    for wrong_type in wrong_types:
        test_value = wrong_type()
        with pytest.raises(SystemExit):
            _check_input_config({key: test_value})


def test_missing_paths():
    """
    If the script is run without any paths, we want to abort.
    """
    with pytest.raises(SystemExit):
        make_config([])



import pytest

from . import v

assert v  # Silence pyflakes.


def test_syntax_error(v):
    v.scan("foo bar")
    assert int(v.report()) == 1


def test_null_byte(v):
    v.scan("\x00")
    assert int(v.report()) == 1


def test_confidence_range(v):
    v.scan(
        """\
def foo():
    pass
"""
    )
    with pytest.raises(ValueError):
        v.get_unused_code(min_confidence=150)



from vulture import core

dc = core.DEFAULT_CONFIDENCE


def check_min_confidence(code, min_confidence, expected):
    v = core.Vulture(verbose=True)
    v.scan(code)
    detected = {
        item.name: item.confidence
        for item in v.get_unused_code(min_confidence=min_confidence)
    }
    assert detected == expected


def test_confidence_import():
    code = """\
import foo
"""
    check_min_confidence(code, 50, {"foo": 90})
    check_min_confidence(code, 100, {})


def test_confidence_unreachable():
    code = """\
def foo():
    return
    bar()

foo()
"""
    check_min_confidence(code, 50, {"return": 100})
    check_min_confidence(code, 100, {"return": 100})


def test_function_arg():
    code = """\
def foo(a):
    b = 3

foo(5)
"""
    check_min_confidence(code, 50, {"a": 100, "b": dc})
    check_min_confidence(code, dc, {"a": 100, "b": dc})
    check_min_confidence(code, 100, {"a": 100})


def test_confidence_class():
    code = """\
class Foo:
    pass
"""
    check_min_confidence(code, 50, {"Foo": dc})
    check_min_confidence(code, 100, {})


def test_confidence_attr():
    code = "A.b = 'something'"
    check_min_confidence(code, 50, {"b": dc})
    check_min_confidence(code, 100, {})


def test_confidence_props():
    code = """\
class Foo:
    @property
    def some_prop():
        pass

Foo()
"""
    check_min_confidence(code, 50, {"some_prop": dc})
    check_min_confidence(code, 100, {})


def test_confidence_async_def():
    code = """\
async def foo():
    if bar():
        pass
    else:
        print("Else")
"""
    check_min_confidence(code, 50, {"foo": dc})
    check_min_confidence(code, 75, {})



import sys

import pytest

from . import check, v

assert v  # Silence pyflakes.


def test_function_object1(v):
    v.scan(
        """\
def func():
    pass

a = func
"""
    )
    check(v.defined_funcs, ["func"])
    check(v.unused_funcs, [])


def test_function_object2(v):
    v.scan(
        """\
def func():
    pass

func
"""
    )
    check(v.defined_funcs, ["func"])
    check(v.unused_funcs, [])


def test_function1(v):
    v.scan(
        """\
def func1(a):
    pass

def func2(b):
    func1(b)
"""
    )
    check(v.defined_funcs, ["func1", "func2"])
    check(v.unused_funcs, ["func2"])


def test_function2(v):
    v.scan(
        """\
def func(a):
    pass

func(5)
"""
    )
    check(v.unused_funcs, [])
    check(v.defined_funcs, ["func"])


def test_function3(v):
    v.scan(
        """\
def foo(a):
    pass

b = foo(5)
"""
    )
    check(v.unused_funcs, [])
    check(v.defined_funcs, ["foo"])


def test_async_function(v):
    v.scan(
        """\
async def foo():
    pass
"""
    )
    check(v.defined_funcs, ["foo"])
    check(v.unused_funcs, ["foo"])


def test_async_method(v):
    v.scan(
        """\
class Foo:
    async def bar(self):
        pass
"""
    )
    check(v.defined_classes, ["Foo"])
    check(v.defined_funcs, [])
    check(v.defined_methods, ["bar"])
    check(v.unused_classes, ["Foo"])
    check(v.unused_methods, ["bar"])


def test_function_and_method1(v):
    v.scan(
        """\
class Bar(object):
    def func(self):
        pass

def func():
    pass

func()
"""
    )
    check(v.defined_classes, ["Bar"])
    check(v.defined_funcs, ["func"])
    check(v.defined_methods, ["func"])
    check(v.unused_classes, ["Bar"])
    check(v.unused_funcs, [])
    # Bar.func is unused, but it's hard to detect this without producing a
    # false positive in test_function_and_method2.
    check(v.unused_methods, [])


def test_function_and_method2(v):
    v.scan(
        """\
class Bar(object):
    def func(self):
        pass

    other_name_for_func = func

Bar().other_name_for_func()
"""
    )
    check(v.defined_classes, ["Bar"])
    check(v.defined_funcs, [])
    check(v.defined_methods, ["func"])
    check(v.defined_vars, ["other_name_for_func"])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_methods, [])
    check(v.unused_vars, [])


def test_attribute1(v):
    v.scan(
        """\
foo.bar = 1
foo.bar = 2
"""
    )
    check(v.unused_funcs, [])
    check(v.defined_funcs, [])
    check(v.defined_attrs, ["bar", "bar"])
    check(v.used_names, ["foo"])
    check(v.unused_attrs, ["bar", "bar"])


def test_ignored_attributes(v):
    v.scan(
        """\
A._ = 0
A._a = 1
A.__b = 2
A.__c__ = 3
A._d_ = 4
"""
    )
    check(v.defined_attrs, ["_", "_a", "__b", "__c__", "_d_"])
    check(v.used_names, ["A"])
    check(v.unused_attrs, ["_", "__b", "__c__", "_a", "_d_"])
    check(v.unused_vars, [])


def test_getattr(v):
    v.scan(
        """\
class Thing:
    used_attr1 = 1
    used_attr2 = 2
    used_attr3 = 3
    unused_attr = 4

getattr(Thing, "used_attr1")
getattr(Thing, "used_attr2", None)
hasattr(Thing, "used_attr3")

# Weird calls ignored
hasattr(Thing, "unused_attr", None)
getattr(Thing)
getattr("unused_attr")
getattr(Thing, "unused_attr", 1, 2)
"""
    )
    check(v.unused_vars, ["unused_attr"])
    check(
        v.used_names,
        [
            "Thing",
            "getattr",
            "hasattr",
            "used_attr1",
            "used_attr2",
            "used_attr3",
        ],
    )


def test_callback1(v):
    v.scan(
        """\
class Bar(object):
    def foo(self):
        pass

b = Bar()
b.foo
"""
    )
    check(v.used_names, ["Bar", "b", "foo"])
    check(v.defined_classes, ["Bar"])
    check(v.defined_funcs, [])
    check(v.defined_methods, ["foo"])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])


def test_class1(v):
    v.scan(
        """\
class Bar(object):
    pass
"""
    )
    check(v.used_names, [])
    check(v.defined_classes, ["Bar"])
    check(v.unused_classes, ["Bar"])


def test_class2(v):
    v.scan(
        """\
class Bar():
    pass
class Foo(Bar):
    pass
Foo()
"""
    )
    check(v.used_names, ["Bar", "Foo"])
    check(v.defined_classes, ["Bar", "Foo"])
    check(v.unused_classes, [])


def test_class3(v):
    v.scan(
        """\
class Bar():
    pass
[Bar]
"""
    )
    check(v.used_names, ["Bar"])
    check(v.defined_classes, ["Bar"])
    check(v.unused_classes, [])


def test_class4(v):
    v.scan(
        """\
class Bar():
    pass
Bar()
"""
    )
    check(v.used_names, ["Bar"])
    check(v.defined_classes, ["Bar"])
    check(v.unused_classes, [])


def test_class5(v):
    v.scan(
        """\
class Bar():
    pass
b = Bar()
"""
    )
    check(v.used_names, ["Bar"])
    check(v.defined_classes, ["Bar"])
    check(v.unused_classes, [])
    check(v.unused_vars, ["b"])


def test_class6(v):
    v.scan(
        """\
class Bar():
    pass
a = []
a.insert(0, Bar())
"""
    )
    check(v.defined_classes, ["Bar"])
    check(v.unused_classes, [])


def test_class7(v):
    v.scan(
        """\
class Bar(object):
    pass

class Foo(object):
    def __init__(self):
        self.b = xyz.Bar(self)
"""
    )
    check(v.defined_classes, ["Bar", "Foo"])
    check(v.unused_classes, ["Foo"])


def test_method1(v):
    v.scan(
        """\
def __init__(self):
    self.a.foo()

class Bar(object):
    def foo(self):
        pass

    @classmethod
    def bar(cls):
        pass

    @staticmethod
    def foobar():
        pass
"""
    )
    check(v.defined_classes, ["Bar"])
    check(v.defined_funcs, [])
    check(v.defined_methods, ["foo", "bar", "foobar"])
    check(v.unused_classes, ["Bar"])
    check(v.unused_funcs, [])
    check(v.unused_methods, ["bar", "foobar"])


def test_token_types(v):
    v.scan(
        """\
a
b = 2
c()
x.d
"""
    )
    check(v.defined_funcs, [])
    check(v.defined_vars, ["b"])
    check(v.used_names, ["a", "c", "d", "x"])
    check(v.unused_attrs, [])
    check(v.unused_funcs, [])
    check(v.unused_props, [])
    check(v.unused_vars, ["b"])


def test_variable1(v):
    v.scan("a = 1\nb = a")
    check(v.defined_funcs, [])
    check(v.used_names, ["a"])
    check(v.defined_vars, ["a", "b"])
    check(v.unused_vars, ["b"])


def test_variable2(v):
    v.scan("a = 1\nc = b.a")
    check(v.defined_funcs, [])
    check(v.defined_vars, ["a", "c"])
    check(v.used_names, ["a", "b"])
    check(v.unused_vars, ["c"])


def test_variable3(v):
    v.scan("(a, b), c = (d, e, f)")
    check(v.defined_funcs, [])
    check(v.defined_vars, ["a", "b", "c"])
    check(v.used_names, ["d", "e", "f"])
    check(v.unused_vars, ["a", "b", "c"])


def test_variable4(v):
    v.scan("for a, b in func(): a")
    check(v.defined_funcs, [])
    check(v.defined_vars, ["a", "b"])
    check(v.used_names, ["a", "func"])
    check(v.unused_vars, ["b"])


def test_variable5(v):
    v.scan("[a for a, b in func()]")
    check(v.defined_vars, ["a", "b"])
    check(v.used_names, ["a", "func"])
    check(v.unused_vars, ["b"])


def test_ignored_variables(v):
    v.scan(
        """\
_ = 0
_a = 1
__b = 2
__c__ = 3
_d_ = 4
"""
    )
    check(v.defined_vars, ["__b"])
    check(sorted(v.used_names), [])
    check(v.unused_vars, ["__b"])


def test_prop1(v):
    v.scan(
        """\
class Bar(object):
    @property
    def prop(self):
        pass

c = Bar()
c.prop
"""
    )
    check(v.defined_classes, ["Bar"])
    check(v.defined_props, ["prop"])
    check(v.unused_classes, [])
    check(v.unused_props, [])


def test_prop2(v):
    v.scan(
        """\
class Bar(object):
    @property
    def prop(self):
        pass

prop = 1
"""
    )
    check(v.defined_classes, ["Bar"])
    check(v.defined_props, ["prop"])
    check(v.defined_vars, ["prop"])
    check(v.unused_classes, ["Bar"])
    check(v.unused_props, ["prop"])


def test_object_attribute(v):
    v.scan(
        """\
class Bar(object):
    def __init__(self):
        self.a = []
"""
    )
    check(v.defined_attrs, ["a"])
    check(v.defined_classes, ["Bar"])
    check(v.defined_vars, [])
    check(v.used_names, [])
    check(v.unused_attrs, ["a"])
    check(v.unused_classes, ["Bar"])


def test_function_names_in_test_file(v):
    v.scan(
        """\
def setup_module(module):
    module

def teardown_module(module):
    module

def setup_function(function):
    function

def teardown_function(function):
    function

def test_func():
    pass

def other_func():
    pass

class TestClass:
    @classmethod
    def setup_class(cls):
        cls

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        method

    def teardown_method(self, method):
        pass

class BasicTestCase:
    pass

class OtherClass:
    pass
""",
        filename="dir/test_function_names.py",
    )
    check(v.defined_attrs, [])
    check(v.defined_classes, ["OtherClass"])
    check(v.defined_funcs, ["other_func"])
    check(v.defined_methods, [])
    check(
        v.defined_vars,
        [
            "cls",
            "cls",
            "function",
            "function",
            "method",
            "method",
            "module",
            "module",
        ],
    )
    check(v.used_names, ["classmethod", "cls", "function", "method", "module"])
    check(v.unused_attrs, [])
    check(v.unused_classes, ["OtherClass"])
    check(v.unused_funcs, ["other_func"])
    check(v.unused_methods, [])
    check(v.unused_vars, [])


def test_async_function_name_in_test_file(v):
    v.scan(
        """\
async def test_func():
    pass

async def other_func():
    pass
""",
        filename="dir/test_function_names.py",
    )
    check(v.defined_funcs, ["other_func"])
    check(v.unused_funcs, ["other_func"])


def test_async_function_name_in_normal_file(v):
    v.scan(
        """\
async def test_func():
    pass

async def other_func():
    pass
""",
        filename="dir/function_names.py",
    )
    check(v.defined_funcs, ["test_func", "other_func"])
    check(v.unused_funcs, ["other_func", "test_func"])


def test_function_names_in_normal_file(v):
    v.scan(
        """\
def test_func():
    pass

def other_func():
    pass

class TestClass:
    pass

class BasicTestCase:
    pass

class OtherClass:
    pass
"""
    )
    check(v.defined_attrs, [])
    check(v.defined_classes, ["BasicTestCase", "OtherClass", "TestClass"])
    check(v.defined_funcs, ["test_func", "other_func"])
    check(v.defined_vars, [])
    check(v.used_names, [])
    check(v.unused_attrs, [])
    check(v.unused_classes, ["BasicTestCase", "OtherClass", "TestClass"])
    check(v.unused_funcs, ["other_func", "test_func"])


def test_global_attribute(v):
    v.scan(
        """\
# Module foo:
a = 1
if a == 1:
    pass

# Module bar:
import foo
foo.a = 2
"""
    )
    check(v.defined_attrs, ["a"])
    check(v.defined_vars, ["a"])
    check(v.used_names, ["a", "foo"])
    check(v.unused_attrs, [])


def test_boolean(v):
    v.scan(
        """\
a = True
a
"""
    )
    check(v.defined_vars, ["a"])
    check(v.used_names, ["a"])
    check(v.unused_vars, [])


def test_builtin_types(v):
    v.scan(
        """\
a = b
a = 1
a = "s"
a = object
a = False
"""
    )
    check(v.defined_vars, ["a"] * 5)
    check(v.used_names, ["b"])
    check(v.unused_vars, ["a"] * 5)


def test_unused_args(v):
    v.scan(
        """\
def foo(x, y):
    return x + 1
"""
    )
    check(v.defined_vars, ["x", "y"])
    check(v.used_names, ["x"])
    check(v.unused_vars, ["y"])


def test_unused_kwargs(v):
    v.scan(
        """\
def foo(x, y=3, **kwargs):
    return x + 1
"""
    )
    check(v.defined_vars, ["kwargs", "x", "y"])
    check(v.used_names, ["x"])
    check(v.unused_vars, ["kwargs", "y"])


def test_unused_kwargs_with_odd_name(v):
    v.scan(
        """\
def foo(**bar):
    pass
"""
    )
    check(v.defined_vars, ["bar"])
    check(v.used_names, [])
    check(v.unused_vars, ["bar"])


def test_unused_vararg(v):
    v.scan(
        """\
def foo(*bar):
    pass
"""
    )
    check(v.defined_vars, ["bar"])
    check(v.used_names, [])
    check(v.unused_vars, ["bar"])


def test_multiple_definition(v):
    v.scan(
        """\
a = 1
a = 2
"""
    )
    check(v.defined_vars, ["a", "a"])
    check(v.used_names, [])
    check(v.unused_vars, ["a", "a"])


def test_arg_type_annotation(v):
    v.scan(
        """\
from typing import Iterable

def f(n: int) -> Iterable[int]:
    yield n
"""
    )

    check(v.unused_vars, [])
    check(v.unused_funcs, ["f"])
    check(v.unused_imports, [])


def test_var_type_annotation(v):
    v.scan(
        """\
from typing import List

x: List[int] = [1]
"""
    )

    check(v.unused_vars, ["x"])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])


def test_type_hint_comments(v):
    v.scan(
        """\
from typing import Any, Dict, List, Text, Tuple


def plain_function(arg):
    # type: (Text) -> None
    pass

async def async_function(arg):
    # type: (List[int]) -> None
    pass

some_var = {}  # type: Dict[str, str]

class Thing:
    def __init__(self):
        self.some_attr = (1, 2)  # type: Tuple[int, int]

for x in []:  # type: Any
    print(x)
"""
    )

    if sys.version_info < (3, 8):
        check(v.unused_imports, ["Any", "Dict", "List", "Text", "Tuple"])
    else:
        check(v.unused_imports, [])
        assert not v.found_dead_code_or_error


def test_invalid_type_comment(v):
    v.scan(
        """\
def bad():
    # type: bogus
    pass
bad()
"""
    )

    if sys.version_info < (3, 8):
        assert not v.found_dead_code_or_error
    else:
        assert v.found_dead_code_or_error


def test_unused_args_with_del(v):
    v.scan(
        """\
def foo(a, b, c, d=3):
    del c, d
    return a + b

foo(1, 2)
"""
    )

    check(v.defined_funcs, ["foo"])
    check(v.defined_vars, ["a", "b", "c", "d"])
    check(v.used_names, ["foo", "a", "b", "c", "d"])
    check(v.unused_vars, [])
    check(v.unused_funcs, [])


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="requires python3.10 or higher"
)
def test_match_class_simple(v):
    v.scan(
        """\
from dataclasses import dataclass


@dataclass
class X:
    a: int
    b: int
    c: int
    u: int

x = input()

match x:
    case X(a=0):
        print("a")
    case X(b=0, c=0):
        print("b c")
"""
    )
    check(v.defined_classes, ["X"])
    check(v.defined_vars, ["a", "b", "c", "u", "x"])

    check(v.unused_classes, [])
    check(v.unused_vars, ["u"])


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="requires python3.10 or higher"
)
def test_match_class_embedded(v):
    v.scan(
        """\
from dataclasses import dataclass


@dataclass
class X:
    a: int
    b: int
    c: int
    d: int
    e: int
    u: int

x = input()

match x:
    case X(a=1) | X(b=0):
        print("Or")
    case [X(c=1), X(d=0)]:
        print("Sequence")
    case {"k": X(e=1)}:
        print("Mapping")
"""
    )
    check(v.defined_classes, ["X"])
    check(v.defined_vars, ["a", "b", "c", "d", "e", "u", "x"])

    check(v.unused_classes, [])
    check(v.unused_vars, ["u"])


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="requires python3.10 or higher"
)
def test_match_enum(v):
    v.scan(
        """\
from enum import Enum


class Color(Enum):
    RED = 0
    YELLOW = 1
    GREEN = 2
    BLUE = 3

color = input()

match color:
    case Color.RED:
        print("Real danger!")
    case Color.YELLOW | Color.GREEN:
        print("No danger!")
"""
    )
    check(v.defined_classes, ["Color"])
    check(v.defined_vars, ["RED", "YELLOW", "GREEN", "BLUE", "color"])

    check(v.unused_classes, [])
    check(v.unused_vars, ["BLUE"])



import ast
import os
import pathlib

import pytest

from vulture import utils


class TestFormatPath:
    @pytest.fixture
    def tmp_cwd(self, tmp_path, monkeypatch):
        cwd = tmp_path / "workingdir"
        cwd.mkdir()
        monkeypatch.chdir(cwd)
        return cwd

    def test_relative_inside(self):
        filepath = pathlib.Path("testfile.py")
        formatted = utils.format_path(filepath)
        assert formatted == filepath
        assert not formatted.is_absolute()

    def test_relative_outside(self, tmp_cwd):
        filepath = pathlib.Path(os.pardir) / "testfile.py"
        formatted = utils.format_path(filepath)
        assert formatted == filepath
        assert not formatted.is_absolute()

    def test_absolute_inside(self, tmp_cwd):
        filepath = tmp_cwd / "testfile.py"
        formatted = utils.format_path(filepath)
        assert formatted == pathlib.Path("testfile.py")
        assert not formatted.is_absolute()

    def test_absolute_outside(self, tmp_cwd):
        filepath = (tmp_cwd / os.pardir / "testfile.py").resolve()
        formatted = utils.format_path(filepath)
        assert formatted == filepath
        assert formatted.is_absolute()


def check_decorator_names(code, expected_names):
    decorator_names = []

    def visit_FunctionDef(node):
        for decorator in node.decorator_list:
            decorator_names.append(utils.get_decorator_name(decorator))

    node_visitor = ast.NodeVisitor()
    node_visitor.visit_AsyncFunctionDef = visit_FunctionDef
    node_visitor.visit_ClassDef = visit_FunctionDef
    node_visitor.visit_FunctionDef = visit_FunctionDef
    node_visitor.visit(ast.parse(code))
    assert expected_names == decorator_names


def test_get_decorator_name_simple():
    code = """\
@foobar
def hoo():
    pass
"""
    check_decorator_names(code, ["@foobar"])


def test_get_decorator_name_call():
    code = """\
@xyz()
def bar():
    pass
"""
    check_decorator_names(code, ["@xyz"])


def test_get_decorator_name_async():
    code = """\
@foo.bar.route('/foobar')
async def async_function(request):
    print(request)
"""
    check_decorator_names(code, ["@foo.bar.route"])


def test_get_decorator_name_multiple_attrs():
    code = """\
@x.y.z
def doo():
    pass
"""
    check_decorator_names(code, ["@x.y.z"])


def test_get_decorator_name_multiple_attrs_called():
    code = """\
@a.b.c.d.foo("Foo and Bar")
def hoofoo():
    pass
"""
    check_decorator_names(code, ["@a.b.c.d.foo"])


def test_get_decorator_name_multiple_decorators():
    code = """\
@foo
@bar()
@x.y.z.a('foobar')
def func():
    pass
"""
    check_decorator_names(code, ["@foo", "@bar", "@x.y.z.a"])


def test_get_decorator_name_class():
    code = """\
@foo
@bar.yz
class Foo:
    pass
"""
    check_decorator_names(code, ["@foo", "@bar.yz"])



from . import check, v

assert v  # Silence pyflakes.


def test_old_format_string(v):
    v.scan("'%(a)s, %(b)d' % locals()")
    check(v.used_names, ["a", "b", "locals"])


def test_new_format_string(v):
    v.scan("'{a}, {b:0d} {c:<30} {d:.2%}'.format(**locals())")
    check(v.used_names, ["a", "b", "c", "d", "format", "locals"])


def test_f_string(v):
    v.scan(
        """\
f'{a}, {b:0d} {c:<30} {d:.2%} {e()} {f:{width}.{precision}}'
f'{ {x:y for (x, y) in ((1, 2), (3, 4))} }'
"""
    )
    check(
        v.used_names,
        ["a", "b", "c", "d", "e", "f", "precision", "width", "x", "y"],
    )


def test_new_format_string_access(v):
    v.scan("'{a.b}, {c.d.e} {f[g]} {h[i][j].k}'.format(**locals())")
    check(
        v.used_names,
        ["a", "b", "c", "d", "e", "f", "h", "k", "format", "locals"],
    )


def test_new_format_string_numbers(v):
    v.scan("'{0.b}, {0.d.e} {0[1]} {0[1][1].k}'.format(**locals())")
    check(v.used_names, ["b", "d", "e", "k", "format", "locals"])


def test_incorrect_format_string(v):
    v.scan('"{"')
    v.scan('"{!-a:}"')
    check(v.used_names, [])


def test_format_string_not_using_locals(v):
    """Strings that are not formatted with locals() should not be parsed."""
    v.scan(
        """\
"{variable}"

def foobar():
    '''
    Return data of the form
        {this_looks_like_a_format_string: 1}
    '''
    pass

"%(thing)s" % {"thing": 1}

"%(apple)s" * locals()

"{} {a} {b}".format(1, a=used_var, b=locals())
    """
    )
    check(v.used_names, ["used_var", "locals", "format"])



import ast

from vulture import lines


def get_last_line_number_slow(node):
    """
    A slower, but also simpler and slightly more accurate version of
    get_last_line_number().

    It traverses all descendants of node and records the highest line
    number seen.

    This code is 1.6 times slower than count_lines() on the Python
    subset of tensorflow. It reports the same sizes for all test cases
    and the functions and classes in tensorflow.

    """
    return max(getattr(node, "lineno", -1) for node in ast.walk(node))


def count_lines(node):
    """Estimate the number of lines of the given AST node."""
    last_lineno = lines.get_last_line_number(node)
    assert get_last_line_number_slow(node) == last_lineno
    return last_lineno - lines.get_first_line_number(node) + 1


def check_size(example, size):
    tree = ast.parse(example)
    for node in tree.body:
        if isinstance(node, ast.ClassDef) and node.name == "Foo":
            assert count_lines(node) == size
            break
    else:
        raise AssertionError('Failed to find top-level class "Foo" in code')


def test_size_basic():
    example = """
class Foo:
    foo = 1
    bar = 2
"""
    check_size(example, 3)


def test_size_class():
    example = """
class Foo(object):
    def bar():
        pass

    @staticmethod
    def func():
        if "foo" == "bar":
            return "xyz"
        import sys
        return len(sys.argv)
"""
    check_size(example, 10)


def test_size_if_else():
    example = """
@identity
class Foo(object):
    @identity
    @identity
    def bar(self):
        if "a" == "b":
            pass
        elif "b" == "c":
            pass
        else:
            pass
"""
    size = 11
    check_size(example, size)


def test_size_decorated_class():
    example = """
@foo
@property
@xoo
class Foo:
    def zoo(self):
        pass
"""
    check_size(example, 6)


def test_size_while():
    example = """
class Foo:
    while 1:
        print(1)
"""
    check_size(example, 3)


def test_size_while_else():
    example = """
class Foo:
    while "b" > "a":
        pass
    else:
        pass
"""
    check_size(example, 5)


def test_size_with():
    example = """
class Foo:
    with open("/dev/null") as f:
        f.write("")
"""
    check_size(example, 3)


def test_size_try_except_else():
    example = """
class Foo:
    try:
        x = sys.argv[99]
    except IndexError:
        pass
    except Exception:
        pass
    else:
        pass
"""
    check_size(example, 9)


def test_size_try_finally():
    example = """
class Foo:
    try:
        1/0
    finally:
        return 99
"""
    check_size(example, 5)


def test_size_try_except():
    example = """
class Foo:
    try:
        foo()
    except:
        bar()
"""
    check_size(example, 5)


def test_size_try_excepts():
    example = """
class Foo:
    try:
        foo()
    except IOError:
        bar()
    except AttributeError:
        pass
"""
    check_size(example, 7)


def test_size_for():
    example = """
class Foo:
    for i in range(10):
        print(i)
"""
    check_size(example, 3)


def test_size_for_else():
    example = """
class Foo:
    for arg in sys.argv:
        print("loop")
    else:
        print("else")
"""
    check_size(example, 5)


def test_size_class_nested():
    example = """
class Foo:
    class Bar:
        pass
"""
    check_size(example, 3)


# We currently cannot handle code ending with multiline strings.
def test_size_multi_line_return():
    example = """
class Foo:
    def foo():
        return (
            'very'
            'long'
            'string')
"""
    check_size(example, 4)


# We currently cannot handle code ending with comment lines.
def test_size_comment_after_last_line():
    example = """
class Foo:
    def bar():
        # A comment.
        pass
        # This comment won't be detected.
"""
    check_size(example, 4)


def test_size_generator():
    example = """
class Foo:
    def bar():
        yield something
"""
    check_size(example, 3)


def test_size_exec():
    example = """
class Foo:
    exec('a')
"""
    check_size(example, 2)


def test_size_print1():
    example = """
class Foo:
    print(
        'foo')
"""
    check_size(example, 3)


def test_size_print2():
    example = """
class Foo:
    print(
        'foo',)
"""
    check_size(example, 3)


def test_size_return():
    example = """
class Foo:
    return (True and
        False)
"""
    check_size(example, 3)


def test_size_import_from():
    example = """
class Foo:
    from a import b
"""
    check_size(example, 2)


def test_size_delete():
    example = """
class Foo:
    del a[:
        foo()]
"""
    check_size(example, 3)


def test_size_list_comprehension():
    example = """
class Foo:
    [a
     for a in
     b]
"""
    check_size(example, 4)


# We currently cannot handle closing brackets on a separate line.
def test_size_list():
    example = """
class Foo:
    [a, b
    ]
"""
    check_size(example, 2)


def test_size_ellipsis():
    example = """
class Foo:
    bar[1:2,
        ...]
"""
    check_size(example, 3)


def test_size_starargs():
    example = """
class Foo:
    def foo():
        bar(*a,
            b=c)
"""
    check_size(example, 4)


# If we add a line break between a and b, the code is too greedy and moves
# down to the slice which has no line numbers. If we took b or c into
# account, the line count would be correct.
def test_size_assign():
    example = """
class Foo:
    bar = foo(a, b)[c,:]
"""
    check_size(example, 2)


def test_size_async_function_def():
    example = """
class Foo:
    async def foo(some_attr):
        pass
"""
    check_size(example, 3)


def test_size_async_with():
    example = """
class Foo:
    async def bar():
        async with x:
            pass
"""
    check_size(example, 4)


def test_size_async_for():
    example = """
class Foo:
    async def foo():
        async for a in b:
            pass
"""
    check_size(example, 4)



import ast
import sys


def _get_last_child_with_lineno(node):
    """
    Return the last direct child of `node` that has a lineno attribute,
    or None if `node` has no such children.

    Almost all node._field lists are sorted by the order in which they
    appear in source code. For some nodes however, we have to skip some
    fields that either don't have line numbers (e.g., "ctx" and "names")
    or that are in the wrong position (e.g., "decorator_list" and
    "returns"). Then we choose the first field (i.e., the field with the
    highest line number) that actually contains a node. If it contains a
    list of nodes, we return the last one.

    """
    ignored_fields = {"ctx", "decorator_list", "names", "returns"}
    fields = node._fields
    # The fields of ast.Call are in the wrong order.
    if isinstance(node, ast.Call):
        fields = ("func", "args", "starargs", "keywords", "kwargs")
    for name in reversed(fields):
        if name in ignored_fields:
            continue

        try:
            last_field = getattr(node, name)
        except AttributeError:
            continue

        # Ignore non-AST objects like "is_async", "level" and "nl".
        if isinstance(last_field, ast.AST):
            return last_field
        elif isinstance(last_field, list) and last_field:
            return last_field[-1]
    return None


def get_last_line_number(node):
    """Estimate last line number of the given AST node.

    The estimate is based on the line number of the last descendant of
    `node` that has a lineno attribute. Therefore, it underestimates the
    size of code ending with, e.g., multiline strings and comments.

    When traversing the tree, we may see a mix of nodes with line
    numbers and nodes without line numbers. We therefore, store the
    maximum line number seen so far and report it at the end. A more
    accurate (but also slower to compute) estimate would traverse all
    children, instead of just the last one, since choosing the last one
    may lead to a path that ends with a node without line number.

    """
    max_lineno = node.lineno
    while True:
        last_child = _get_last_child_with_lineno(node)
        if last_child is None:
            return max_lineno
        else:
            try:
                max_lineno = max(max_lineno, last_child.lineno)
            except AttributeError:
                pass
        node = last_child


def get_first_line_number(node):
    """
    From Python 3.8 onwards, lineno for decorated objects is the line at which
    the object definition starts, which is different from what Python < 3.8
    reported -- the lineno of the first decorator. To preserve this behaviour
    of Vulture for newer Python versions, which is also more accurate for
    counting the size of the unused code chunk (if the property is unused, we
    also don't need it's decorators), we return the lineno of the first
    decorator, if there are any.
    """
    if sys.version_info >= (3, 8):
        decorators = getattr(node, "decorator_list", [])
        if decorators:
            return decorators[0].lineno
    return node.lineno



import pytest

from . import check, v

assert v  # silence pyflakes


@pytest.fixture
def check_whitelist(v):
    def examine(code, results_before, results_after):
        v.scan(code)
        check(v.get_unused_code(), results_before)
        for item in v.get_unused_code():
            v.scan(item.get_whitelist_string())
        check(v.get_unused_code(), results_after)

    return examine


def test_unused_function(check_whitelist):
    code = """\
def func():
    pass
"""
    check_whitelist(code, ["func"], [])


def test_unused_class(check_whitelist):
    code = """\
class Foo:
    def __init__(self):
        pass
"""
    check_whitelist(code, ["Foo"], [])


def test_unused_variables(check_whitelist):
    code = """\
foo = 'unused'
bar = 'variable'
"""
    check_whitelist(code, ["foo", "bar"], [])


def test_unused_import(check_whitelist):
    code = """\
import xyz
import foo as bar
from abc import iou
from lorem import ipsum as dolor
"""
    check_whitelist(code, ["xyz", "bar", "iou", "dolor"], [])


def test_unused_attribute(check_whitelist):
    code = """\
class Foo:
    def bar(self):
        self.foobar = 'unused attr'
"""
    check_whitelist(code, ["Foo", "bar", "foobar"], [])


def test_unused_property(check_whitelist):
    code = """\
class Foo:
    @property
    def bar(self):
        pass
"""
    check_whitelist(code, ["Foo", "bar"], [])


def test_unreachable_code(check_whitelist):
    code = """\
def foo():
    return "Foo Bar"
    print("Hello")
"""
    check_whitelist(code, ["foo", "return"], ["return"])



from . import v

assert v  # Silence pyflakes


def test_sorting(v):
    v.scan(
        """\
def foo():
    print("Hello, I am a long function.")
    return "World"

def bar():
    pass
"""
    )
    assert [item.name for item in v.get_unused_code(sort_by_size=True)] == [
        "bar",
        "foo",
    ]
    assert [item.name for item in v.get_unused_code(sort_by_size=False)] == [
        "foo",
        "bar",
    ]



from vulture import core

from . import check


def check_ignore(code, ignore_names, ignore_decorators, expected):
    v = core.Vulture(
        verbose=True,
        ignore_names=ignore_names,
        ignore_decorators=ignore_decorators,
    )
    v.scan(code)
    check(v.get_unused_code(), expected)


def test_var():
    code = """\
fio = 1
fao = 2
bar = 2
ftobar = 3
baz = 10000
funny = True
"""
    check_ignore(code, ["f?o*", "ba[rz]"], [], ["funny"])


def test_function():
    code = """\
def foo_one():
    pass
def foo_two():
    pass
def foo():
    pass
def bar():
    pass
"""
    check_ignore(code, ["foo*"], [], ["bar"])


def test_async_function():
    code = """\
async def foobar():
    pass
async def bar():
    pass
"""
    check_ignore(code, ["foo*"], [], ["bar"])


def test_class():
    code = """\
class Foo:
    def __init__(self):
        pass
"""
    check_ignore(code, ["Foo"], [], [])


def test_class_ignore():
    code = """\
@bar
class Foo:
    pass

class Bar:
    pass
"""
    check_ignore(code, [], [], ["Foo", "Bar"])


def test_property():
    code = """\
class Foo:
    @property
    def some_property(self, a):
        return a

    @property
    @bar
    def foo_bar(self):
        return 'bar'
"""
    check_ignore(code, ["Foo"], ["@property"], [])
    check_ignore(code, ["Foo"], [], ["some_property", "foo_bar"])


def test_attribute():
    code = """\
class Foo:
    def __init__(self, attr_foo, attr_bar):
        self._attr_foo = attr_foo
        self._attr_bar = attr_bar
"""
    check_ignore(code, ["foo", "*_foo"], [], ["Foo", "_attr_bar"])


def test_decorated_functions():
    code = """\
def decor():
    return help

class FooBar:
    def foobar(self):
        return help

    @property
    def prop_one(self):
        pass

f = FooBar()

@decor()
def bar():
    pass

@f.foobar
def foo():
    pass

@bar
@foo
@f.foobar()
def barfoo():
    pass
"""
    check_ignore(code, [], ["@decor", "*@f.foobar"], ["prop_one"])
    check_ignore(code, [], ["*decor", "@*f.foobar"], ["prop_one"])


def test_decorated_async_functions():
    code = """\
@app.route('something')
@foobar
async def async_function():
    pass

@a.b.c
async def foo():
    pass
"""
    check_ignore(code, [], ["@app.route", "@a.b"], ["foo"])


def test_decorated_property():
    code = """\
@bar
@property
def foo():
    pass
"""
    check_ignore(code, [], ["@bar"], [])
    check_ignore(code, [], ["@baz"], ["foo"])
    check_ignore(code, [], ["@property"], [])


def test_decorated_property_reversed():
    code = """\
@property
@bar
def foo():
    pass
"""
    check_ignore(code, [], ["@bar"], [])
    check_ignore(code, [], ["@property"], [])
    check_ignore(code, [], ["@b*r"], [])
    check_ignore(code, [], ["@barfoo"], ["foo"])


def test_decorated_class():
    code = """\
@barfoo
@foo.bar('foo')
class Bar:
    def __init__(self):
        pass
"""
    check_ignore(code, [], [], ["Bar"])
    check_ignore(code, [], ["@bar*"], [])



from . import check, v

assert v  # Silence pyflakes.


def test_import_star(v):
    v.scan(
        """\
from a import *
from a.b import *
"""
    )
    check(v.defined_imports, [])
    check(v.unused_imports, [])


def test_import_from_future(v):
    v.scan("""from __future__ import division""")
    check(v.defined_imports, [])
    check(v.unused_imports, [])


def test_double_import(v):
    v.scan(
        """\
import foo as bar
import foo
"""
    )
    check(v.defined_imports, ["bar", "foo"])
    # Once the bar import is removed, the foo import will be detected.
    check(v.unused_imports, ["bar"])


def test_attribute_access(v):
    v.scan(
        """\
# foo.py
class Foo:
    pass

# bar.py
from foo import Foo

# main.py
import bar
bar.Foo
"""
    )
    check(v.defined_imports, ["Foo", "bar"])
    check(v.unused_imports, [])


def test_nested_import(v):
    v.scan(
        """\
import os.path
os.path.expanduser("~")
"""
    )
    check(v.defined_imports, ["os"])
    check(v.used_names, ["os", "path", "expanduser"])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])


definitions = """\
class A(object):
    pass
class B(object):
    pass
def C():
    pass
D = 42
"""

imports = """\
from any_module import A
import B, C
import D
"""

aliased_imports = """\
from any_module import A as AliasA
import B as AliasB, C as AliasC
import D as AliasD
"""

uses = """\
A()
B()
C()
D()
"""

aliased_uses = """\
AliasA()
AliasB()
AliasC()
AliasD()
"""


def test_definitions(v):
    v.scan(definitions)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, [])
    check(v.defined_vars, ["D"])
    check(v.used_names, [])
    check(v.unused_classes, ["A", "B"])
    check(v.unused_funcs, ["C"])
    check(v.unused_imports, [])
    check(v.unused_vars, ["D"])


def test_use_original(v):
    v.scan(definitions + uses)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, [])
    check(v.defined_vars, ["D"])
    check(v.used_names, ["A", "B", "C", "D"])
    check(v.unused_funcs, [])
    check(v.unused_classes, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])


def test_import_original(v):
    v.scan(definitions + imports)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, ["A", "B", "C", "D"])
    check(v.defined_vars, ["D"])
    check(v.used_names, [])
    check(v.unused_classes, ["A", "B"])
    check(v.unused_funcs, ["C"])
    check(v.unused_imports, ["A", "B", "C", "D"])
    check(v.unused_vars, ["D"])


def test_import_original_use_original(v):
    v.scan(definitions + imports + uses)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, ["A", "B", "C", "D"])
    check(v.defined_vars, ["D"])
    check(v.used_names, ["A", "B", "C", "D"])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])


def test_import_original_use_alias(v):
    v.scan(definitions + imports + aliased_uses)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, ["A", "B", "C", "D"])
    check(v.defined_vars, ["D"])
    check(v.used_names, ["AliasA", "AliasB", "AliasC", "AliasD"])
    check(v.unused_classes, ["A", "B"])
    check(v.unused_funcs, ["C"])
    check(v.unused_imports, ["A", "B", "C", "D"])
    check(v.unused_vars, ["D"])


def test_import_alias(v):
    v.scan(definitions + aliased_imports)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, ["AliasA", "AliasB", "AliasC", "AliasD"])
    check(v.defined_vars, ["D"])
    check(v.used_names, ["A", "B", "C", "D"])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, ["AliasA", "AliasB", "AliasC", "AliasD"])
    check(v.unused_vars, [])


def test_import_alias_use_original(v):
    v.scan(definitions + aliased_imports + uses)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, ["AliasA", "AliasB", "AliasC", "AliasD"])
    check(v.defined_vars, ["D"])
    check(v.used_names, ["A", "B", "C", "D"])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, ["AliasA", "AliasB", "AliasC", "AliasD"])
    check(v.unused_vars, [])


def test_import_alias_use_alias(v):
    v.scan(definitions + aliased_imports + aliased_uses)
    check(v.defined_classes, ["A", "B"])
    check(v.defined_funcs, ["C"])
    check(v.defined_imports, ["AliasA", "AliasB", "AliasC", "AliasD"])
    check(v.defined_vars, ["D"])
    check(
        v.used_names,
        ["A", "B", "C", "D", "AliasA", "AliasB", "AliasC", "AliasD"],
    )
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])


def test_import_with__all__(v):
    v.scan(
        """\
# define.py
class Foo:
    pass

class Bar:
    pass

# main.py
from define import Foo, Bar

__all__ = ["Foo"]

"""
    )
    check(v.defined_imports, ["Foo", "Bar"])
    check(v.unused_imports, ["Bar"])


def test_import_with__all__normal_reference(v):
    v.scan(
        """\
# define.py
class Foo:
    pass

class Bar:
    pass

# main.py
from define import Foo, Bar

__all__ = [Foo]

"""
    )
    check(v.defined_imports, ["Foo", "Bar"])
    check(v.unused_imports, ["Bar"])


def test_import_with__all__string(v):
    v.scan(
        """\
# define.py
class Foo:
    pass

class Bar:
    pass

# main.py
from define import Foo, Bar

__all__ = "Foo"

"""
    )
    check(v.defined_imports, ["Foo", "Bar"])
    # __all__ is not a list or tuple, so Foo is unused.
    check(v.unused_imports, ["Foo", "Bar"])


def test_import_with__all__assign_other_module(v):
    v.scan(
        """\
# define.py
class Foo:
    pass

class Bar:
    pass

# main.py
import define
from define import Foo, Bar

define.__all__ = ["Foo"]

"""
    )
    check(v.defined_imports, ["define", "Foo", "Bar"])
    # Only assignments to __all__ of the current module are covered.
    check(v.unused_imports, ["Foo", "Bar"])


def test_ignore_init_py_files(v):
    v.scan(
        """\
import bar
from foo import *
from zoo import zebra

unused_var = 'monty'
""",
        filename="nested/project/__init__.py",
    )
    check(v.unused_imports, [])
    check(v.unused_vars, ["unused_var"])



import pytest

from vulture.core import ERROR_CODES
from vulture.noqa import NOQA_CODE_MAP, NOQA_REGEXP, _parse_error_codes
from . import check, v

assert v  # Silence pyflakes.


@pytest.mark.parametrize(
    "line, codes",
    [
        ("# noqa", ["all"]),
        ("## noqa", ["all"]),
        ("# noqa Hi, go on.", ["all"]),
        ("# noqa: V101", ["V101"]),
        ("# noqa: V101, V106", ["V101", "V106"]),
        ("# NoQA: V101,      V103, \t V104", ["V101", "V103", "V104"]),
    ],
)
def test_noqa_regex_present(line, codes):
    match = NOQA_REGEXP.search(line)
    parsed = _parse_error_codes(match)
    assert parsed == codes


@pytest.mark.parametrize(
    "line",
    [
        ("# noqa: 123V"),
        ("# noqa explanation: V012"),
        ("# noqa: ,V101"),
        ("# noqa: #noqa: V102"),
        ("# noqa: # noqa: V102"),
    ],
)
def test_noqa_regex_no_groups(line):
    assert NOQA_REGEXP.search(line).groupdict()["codes"] is None


@pytest.mark.parametrize(
    "line",
    [("#noqa"), ("##noqa"), ("# n o q a"), ("#NOQA"), ("# Hello, noqa")],
)
def test_noqa_regex_not_present(line):
    assert not NOQA_REGEXP.search(line)


def test_noqa_without_codes(v):
    v.scan(
        """\
import this  # noqa

@underground  # noqa
class Cellar:
    @property  # noqa
    def wine(self):
        grapes = True  # noqa

    @without_ice  # noqa
    def serve(self, quantity=50):
        self.quantity_served = quantity  # noqa
        return
        self.pour()  # noqa
"""
    )
    check(v.unused_attrs, [])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_props, [])
    check(v.unreachable_code, [])
    check(v.unused_vars, [])


def test_noqa_specific_issue_codes(v):
    v.scan(
        """\
import this  # noqa: V104

@underground  # noqa: V102
class Cellar:
    @property  # noqa: V106
    def wine(self):
        grapes = True  # noqa: V107

    @without_ice  # noqa: V103
    def serve(self, quantity=50):
        self.quantity_served = quantity  # noqa: V101
        return
        self.pour()  # noqa: V201
"""
    )
    check(v.unused_attrs, [])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_methods, ["serve"])
    check(v.unused_props, [])
    check(v.unreachable_code, [])
    check(v.unused_vars, [])


def test_noqa_attributes(v):
    v.scan(
        """\
something.x = 'x'  # noqa: V101
something.z = 'z'  # noqa: V107 (code for unused variable)
something.u = 'u'  # noqa
"""
    )
    check(v.unused_attrs, ["z"])


def test_noqa_classes(v):
    v.scan(
        """\
class QtWidget:  # noqa: V102
    pass

class ABC(QtWidget):
    pass  # noqa: V102 (should not ignore)

class DEF:  # noqa
    pass
"""
    )
    check(v.unused_classes, ["ABC"])


def test_noqa_functions(v):
    v.scan(
        """\
def play(tune, instrument='bongs', _hz='50'):  # noqa: V103
    pass


# noqa
def problems():  # noqa: V104
    pass  # noqa: V103

def hello(name):  # noqa
    print("Hello")
"""
    )
    check(v.unused_funcs, ["problems"])
    check(v.unused_vars, ["instrument", "tune"])


def test_noqa_imports(v):
    v.scan(
        """\
import foo
import this  # noqa: V104
import zoo
from koo import boo  # noqa
from me import *
import dis  # noqa: V101 (code for unused attr)
"""
    )
    check(v.unused_imports, ["foo", "zoo", "dis"])


def test_noqa_properties(v):
    v.scan(
        """\
class Zoo:
    @property
    def no_of_koalas(self):  # noqa
        pass

    @property
    def area(self, width, depth):  # noqa: V105
        pass

    @property  # noqa
    def entry_gates(self):
        pass

    @property  # noqa: V103 (code for unused function)
    def tickets(self):
        pass
"""
    )
    check(v.unused_props, ["no_of_koalas", "area", "tickets"])
    check(v.unused_classes, ["Zoo"])
    check(v.unused_vars, ["width", "depth"])


def test_noqa_multiple_decorators(v):
    v.scan(
        """\
@bar  # noqa: V102
class Foo:
    @property  # noqa: V106
    @make_it_cool
    @log
    def something(self):
        pass

    @coolify
    @property
    def something_else(self):  # noqa: V106
        pass

    @a
    @property
    @b  # noqa
    def abcd(self):
        pass
"""
    )
    check(v.unused_props, ["something_else", "abcd"])
    check(v.unused_classes, [])


def test_noqa_unreacahble_code(v):
    v.scan(
        """\
def shave_sheep(sheep):
    for a_sheep in sheep:
        if a_sheep.is_bald:
            continue
            a_sheep.grow_hair()  # noqa: V201
        a_sheep.shave()
    return
    for a_sheep in sheep:  # noqa: V201
        if a_sheep.still_has_hair:
            a_sheep.shave_again()
"""
    )
    check(v.unreachable_code, [])
    check(v.unused_funcs, ["shave_sheep"])


def test_noqa_variables(v):
    v.scan(
        """\
mitsi = "Mother"  # noqa: V107
harry = "Father"  # noqa
shero = "doggy"   # noqa: V101, V104 (code for unused import, attr)
shinchan.friend = ['masao']  # noqa: V107 (code for unused variable)
"""
    )
    check(v.unused_vars, ["shero"])
    check(v.unused_attrs, ["friend"])


def test_noqa_with_multiple_issue_codes(v):
    v.scan(
        """\
def world(axis):  # noqa: V103, V201
    pass


for _ in range(3):
    continue
    xyz = hello(something, else):  # noqa: V201, V107
"""
    )
    check(v.get_unused_code(), [])


def test_noqa_on_empty_line(v):
    v.scan(
        """\
# noqa
import this
# noqa
"""
    )
    check(v.unused_imports, ["this"])


def test_noqa_with_invalid_codes(v):
    v.scan(
        """\
import this  # V098, A123, F876
"""
    )
    check(v.unused_imports, ["this"])


@pytest.mark.parametrize(
    "first_file, second_file",
    [
        ("foo = None", "bar = None  # noqa"),
        ("bar = None  # noqa", "foo = None"),
    ],
)
def test_noqa_multiple_files(first_file, second_file, v):
    v.scan(first_file, filename="first_file.py")
    v.scan(second_file, filename="second_file.py")
    check(v.unused_vars, ["foo"])


def test_flake8_noqa_codes(v):
    assert NOQA_CODE_MAP["F401"] == ERROR_CODES["import"]
    assert NOQA_CODE_MAP["F841"] == ERROR_CODES["variable"]
    v.scan(
        """\
import this  # noqa: F401

def foo():
    bar = 2  # noqa: F841
"""
    )
    check(v.unused_funcs, ["foo"])
    check(v.unused_imports, [])
    check(v.unused_vars, [])



from . import v

assert v  # Silence pyflakes


def test_item_repr(v):
    v.scan(
        """\
import os

message = "foobar"

class Foo:
    def bar():
        pass
"""
    )
    for item in v.get_unused_code():
        assert repr(item) == f"'{item.name}'"


def test_item_attr(v):
    v.scan("foo.bar = 'bar'")
    assert len(v.unused_attrs) == 1
    a = v.unused_attrs[0]
    assert a.name == "bar"
    assert a.first_lineno == 1
    assert a.last_lineno == 1


def test_item_class(v):
    v.scan(
        """\
class Foo:
    pass
"""
    )
    assert len(v.unused_classes) == 1
    c = v.unused_classes[0]
    assert c.name == "Foo"
    assert c.first_lineno == 1
    assert c.last_lineno == 2


def test_item_function(v):
    v.scan(
        """\
def add(a, b):
    return a + b
"""
    )
    assert len(v.unused_funcs) == 1
    f = v.unused_funcs[0]
    assert f.name == "add"
    assert f.first_lineno == 1
    assert f.last_lineno == 2


def test_item_import(v):
    v.scan(
        """\
import bar
from foo import *
"""
    )
    assert len(v.unused_imports) == 1
    i = v.unused_imports[0]
    assert i.name == "bar"
    assert i.first_lineno == 1
    assert i.last_lineno == 1


def test_item_property(v):
    v.scan(
        """\
@awesomify
class Foo:
    @property
    @wifi(
        username='dog',
        password='cat',
    )
    def bar(self):
        pass
Foo()
"""
    )
    assert len(v.unused_props) == 1
    p = v.unused_props[0]
    assert p.name == "bar"
    assert p.first_lineno == 3
    assert p.last_lineno == 9


def test_item_variable(v):
    v.scan("v = 'Vulture'")
    assert len(v.unused_vars) == 1
    var = v.unused_vars[0]
    assert var.name == "v"
    assert var.first_lineno == 1
    assert var.last_lineno == 1



import ast
import sys

from vulture import utils

from . import check_unreachable
from . import v

assert v  # Silence pyflakes


def check_condition(code, result):
    condition = ast.parse(code, mode="eval").body
    if result:
        assert utils.condition_is_always_true(condition)
    else:
        assert utils.condition_is_always_false(condition)


def test_false():
    check_condition("False", False)
    check_condition("None", False)
    check_condition("0", False)
    # Only Python 3.0-3.6 allows addition and subtraction in ast.literal_eval.
    # (see https://bugs.python.org/issue31778)
    if (3, 0) <= sys.version_info < (3, 7):
        check_condition("1 - 1", False)


def test_empty():
    check_condition("''", False)
    check_condition("[]", False)
    check_condition("{}", False)


def test_true():
    check_condition("True", True)
    check_condition("2", True)
    check_condition("'s'", True)
    check_condition("['foo', 'bar']", True)
    check_condition("{'a': 1, 'b': 2}", True)


def test_complex_conditions():
    conditions = [
        ("foo and False", True, False),
        ("foo or False", False, False),
        ("foo and True", False, False),
        ("foo or True", False, True),
        ("False and foo", True, False),
        ("False and 1", True, False),
        ("not False", False, True),
        ("not True", True, False),
        ("not foo", False, False),
        ("foo and (False or [])", True, False),
        ('(foo and bar) or {"a": 1}', False, True),
    ]
    for condition, always_false, always_true in conditions:
        condition = ast.parse(condition, mode="eval").body
        assert not (always_false and always_true)
        assert utils.condition_is_always_false(condition) == always_false
        assert utils.condition_is_always_true(condition) == always_true


def test_errors():
    conditions = [
        "foo",
        '__name__ == "__main__"',
        "chr(-1)",
        'getattr(True, "foo")',
        'hasattr(str, "foo")',
        "isinstance(True, True)",
        "globals()",
        "locals()",
        "().__class__",
    ]
    for condition in conditions:
        condition = ast.parse(condition, mode="eval").body
        assert not utils.condition_is_always_false(condition)
        assert not utils.condition_is_always_true(condition)


def test_while(v):
    v.scan(
        """\
while False:
    pass
"""
    )
    check_unreachable(v, 1, 2, "while")


def test_while_nested(v):
    v.scan(
        """\
while True:
    while False:
        pass
"""
    )
    check_unreachable(v, 2, 2, "while")


def test_if_false(v):
    v.scan(
        """\
if False:
    pass
"""
    )
    check_unreachable(v, 1, 2, "if")


def test_elif_false(v):
    v.scan(
        """\
if bar():
    pass
elif False:
    print("Unreachable")
"""
    )
    check_unreachable(v, 3, 2, "if")


def test_nested_if_statements_false(v):
    v.scan(
        """\
if foo():
    if bar():
        pass
    elif False:
        print("Unreachable")
        pass
    elif something():
        print("Reachable")
    else:
        pass
else:
    pass
"""
    )
    check_unreachable(v, 4, 3, "if")


def test_if_false_same_line(v):
    v.scan(
        """\
if False: a = 1
else: c = 3
"""
    )
    check_unreachable(v, 1, 1, "if")


def test_if_true(v):
    v.scan(
        """\
if True:
    a = 1
    b = 2
else:
    c = 3
    d = 3
"""
    )
    # For simplicity, we don't report the "else" line as dead code.
    check_unreachable(v, 5, 2, "else")


def test_if_true_same_line(v):
    v.scan(
        """\
if True:
    a = 1
    b = 2
else: c = 3
d = 3
"""
    )
    check_unreachable(v, 4, 1, "else")


def test_nested_if_statements_true(v):
    v.scan(
        """\
if foo():
    if bar():
        pass
    elif True:
        if something():
            pass
        else:
            pass
    elif something_else():
        print("foo")
    else:
        print("bar")
else:
    pass
"""
    )
    check_unreachable(v, 9, 4, "else")


def test_redundant_if(v):
    v.scan(
        """\
if [5]:
    pass
"""
    )
    print(v.unreachable_code[0].size)
    check_unreachable(v, 1, 2, "if")


def test_if_exp_true(v):
    v.scan("foo if True else bar")
    check_unreachable(v, 1, 1, "ternary")


def test_if_exp_false(v):
    v.scan("foo if False else bar")
    check_unreachable(v, 1, 1, "ternary")



import glob
import os.path
import subprocess
import sys

from . import call_vulture, REPO, WHITELISTS


def test_module_with_explicit_whitelists():
    assert call_vulture(["vulture/"] + WHITELISTS) == 0


def test_module_with_implicit_whitelists():
    assert call_vulture(["vulture/"]) == 0


def test_module_without_whitelists():
    assert call_vulture(["vulture/", "--exclude", "whitelists"]) == 1


def test_missing_file():
    assert call_vulture(["missing.py"]) == 1


def test_tests():
    assert call_vulture(["tests/"]) == 0


def test_whitelists_with_python():
    for whitelist in WHITELISTS:
        assert subprocess.call([sys.executable, whitelist], cwd=REPO) == 0


def test_pyc():
    assert call_vulture(["missing.pyc"]) == 1


def test_sort_by_size():
    assert call_vulture(["vulture/utils.py", "--sort-by-size"]) == 1


def test_min_confidence():
    assert (
        call_vulture(
            [
                "vulture/core.py",
                "--exclude",
                "whitelists",
                "--min-confidence",
                "100",
            ]
        )
        == 0
    )


def test_exclude():
    def get_csv(paths):
        return ",".join(os.path.join("vulture", path) for path in paths)

    def call_vulture_with_excludes(excludes):
        return call_vulture(["vulture/", "--exclude", get_csv(excludes)])

    assert call_vulture_with_excludes(["core.py", "utils.py"]) == 1
    assert call_vulture_with_excludes(glob.glob("vulture/*.py")) == 0


def test_make_whitelist():
    assert (
        call_vulture(
            ["vulture/", "--make-whitelist", "--exclude", "whitelists"]
        )
        == 1
    )
    assert call_vulture(["vulture/", "--make-whitelist"]) == 0



import ast
from fnmatch import fnmatch, fnmatchcase
from pathlib import Path
import pkgutil
import re
import string
import sys

from vulture import lines
from vulture import noqa
from vulture import utils
from vulture.config import make_config


DEFAULT_CONFIDENCE = 60

IGNORED_VARIABLE_NAMES = {"object", "self"}
PYTEST_FUNCTION_NAMES = {
    "setup_module",
    "teardown_module",
    "setup_function",
    "teardown_function",
}
PYTEST_METHOD_NAMES = {
    "setup_class",
    "teardown_class",
    "setup_method",
    "teardown_method",
}

ERROR_CODES = {
    "attribute": "V101",
    "class": "V102",
    "function": "V103",
    "import": "V104",
    "method": "V105",
    "property": "V106",
    "variable": "V107",
    "unreachable_code": "V201",
}


def _get_unused_items(defined_items, used_names):
    unused_items = [
        item for item in set(defined_items) if item.name not in used_names
    ]
    unused_items.sort(key=lambda item: item.name.lower())
    return unused_items


def _is_special_name(name):
    return name.startswith("__") and name.endswith("__")


def _match(name, patterns, case=True):
    func = fnmatchcase if case else fnmatch
    return any(func(name, pattern) for pattern in patterns)


def _is_test_file(filename):
    return _match(
        filename.resolve(),
        ["*/test/*", "*/tests/*", "*/test*.py", "*[-_]test.py"],
        case=False,
    )


def _assigns_special_variable__all__(node):
    assert isinstance(node, ast.Assign)
    return isinstance(node.value, (ast.List, ast.Tuple)) and any(
        target.id == "__all__"
        for target in node.targets
        if isinstance(target, ast.Name)
    )


def _ignore_class(filename, class_name):
    return _is_test_file(filename) and "Test" in class_name


def _ignore_import(filename, import_name):
    """
    Ignore star-imported names since we can't detect whether they are used.
    Ignore imports from __init__.py files since they're commonly used to
    collect objects from a package.
    """
    return filename.name == "__init__.py" or import_name == "*"


def _ignore_function(filename, function_name):
    return (
        function_name in PYTEST_FUNCTION_NAMES
        or function_name.startswith("test_")
    ) and _is_test_file(filename)


def _ignore_method(filename, method_name):
    return _is_special_name(method_name) or (
        (method_name in PYTEST_METHOD_NAMES or method_name.startswith("test_"))
        and _is_test_file(filename)
    )


def _ignore_variable(filename, varname):
    """
    Ignore _ (Python idiom), _x (pylint convention) and
    __x__ (special variable or method), but not __x.
    """
    return (
        varname in IGNORED_VARIABLE_NAMES
        or (varname.startswith("_") and not varname.startswith("__"))
        or _is_special_name(varname)
    )


class Item:
    """
    Hold the name, type and location of defined code.
    """

    __slots__ = (
        "name",
        "typ",
        "filename",
        "first_lineno",
        "last_lineno",
        "message",
        "confidence",
    )

    def __init__(
        self,
        name,
        typ,
        filename,
        first_lineno,
        last_lineno,
        message="",
        confidence=DEFAULT_CONFIDENCE,
    ):
        self.name = name
        self.typ = typ
        self.filename = filename
        self.first_lineno = first_lineno
        self.last_lineno = last_lineno
        self.message = message or f"unused {typ} '{name}'"
        self.confidence = confidence

    @property
    def size(self):
        assert self.last_lineno >= self.first_lineno
        return self.last_lineno - self.first_lineno + 1

    def get_report(self, add_size=False):
        if add_size:
            line_format = "line" if self.size == 1 else "lines"
            size_report = f", {self.size:d} {line_format}"
        else:
            size_report = ""
        return "{}:{:d}: {} ({}% confidence{})".format(
            utils.format_path(self.filename),
            self.first_lineno,
            self.message,
            self.confidence,
            size_report,
        )

    def get_whitelist_string(self):
        filename = utils.format_path(self.filename)
        if self.typ == "unreachable_code":
            return f"# {self.message} ({filename}:{self.first_lineno})"
        else:
            prefix = ""
            if self.typ in ["attribute", "method", "property"]:
                prefix = "_."
            return "{}{}  # unused {} ({}:{:d})".format(
                prefix, self.name, self.typ, filename, self.first_lineno
            )

    def _tuple(self):
        return (self.filename, self.first_lineno, self.name)

    def __repr__(self):
        return repr(self.name)

    def __eq__(self, other):
        return self._tuple() == other._tuple()

    def __hash__(self):
        return hash(self._tuple())


class Vulture(ast.NodeVisitor):
    """Find dead code."""

    def __init__(
        self, verbose=False, ignore_names=None, ignore_decorators=None
    ):
        self.verbose = verbose

        def get_list(typ):
            return utils.LoggingList(typ, self.verbose)

        self.defined_attrs = get_list("attribute")
        self.defined_classes = get_list("class")
        self.defined_funcs = get_list("function")
        self.defined_imports = get_list("import")
        self.defined_methods = get_list("method")
        self.defined_props = get_list("property")
        self.defined_vars = get_list("variable")
        self.unreachable_code = get_list("unreachable_code")

        self.used_names = utils.LoggingSet("name", self.verbose)

        self.ignore_names = ignore_names or []
        self.ignore_decorators = ignore_decorators or []

        self.filename = Path()
        self.code = []
        self.found_dead_code_or_error = False

    def scan(self, code, filename=""):
        filename = Path(filename)
        self.code = code.splitlines()
        self.noqa_lines = noqa.parse_noqa(self.code)
        self.filename = filename

        def handle_syntax_error(e):
            text = f' at "{e.text.strip()}"' if e.text else ""
            print(
                f"{utils.format_path(filename)}:{e.lineno}: {e.msg}{text}",
                file=sys.stderr,
            )
            self.found_dead_code_or_error = True

        try:
            node = (
                ast.parse(
                    code, filename=str(self.filename), type_comments=True
                )
                if sys.version_info >= (3, 8)  # type_comments requires 3.8+
                else ast.parse(code, filename=str(self.filename))
            )
        except SyntaxError as err:
            handle_syntax_error(err)
        except ValueError as err:
            # ValueError is raised if source contains null bytes.
            print(
                f'{utils.format_path(filename)}: invalid source code "{err}"',
                file=sys.stderr,
            )
            self.found_dead_code_or_error = True
        else:
            # When parsing type comments, visiting can throw SyntaxError.
            try:
                self.visit(node)
            except SyntaxError as err:
                handle_syntax_error(err)

    def scavenge(self, paths, exclude=None):
        def prepare_pattern(pattern):
            if not any(char in pattern for char in "*?["):
                pattern = f"*{pattern}*"
            return pattern

        exclude = [prepare_pattern(pattern) for pattern in (exclude or [])]

        def exclude_path(path):
            return _match(path, exclude, case=False)

        paths = [Path(path) for path in paths]

        for module in utils.get_modules(paths):
            if exclude_path(module):
                self._log("Excluded:", module)
                continue

            self._log("Scanning:", module)
            try:
                module_string = utils.read_file(module)
            except utils.VultureInputException as err:  # noqa: F841
                print(
                    f"Error: Could not read file {module} - {err}\n"
                    f"Try to change the encoding to UTF-8.",
                    file=sys.stderr,
                )
                self.found_dead_code_or_error = True
            else:
                self.scan(module_string, filename=module)

        unique_imports = {item.name for item in self.defined_imports}
        for import_name in unique_imports:
            path = Path("whitelists") / (import_name + "_whitelist.py")
            if exclude_path(path):
                self._log("Excluded whitelist:", path)
            else:
                try:
                    module_data = pkgutil.get_data("vulture", str(path))
                    self._log("Included whitelist:", path)
                except OSError:
                    # Most imported modules don't have a whitelist.
                    continue
                module_string = module_data.decode("utf-8")
                self.scan(module_string, filename=path)

    def get_unused_code(self, min_confidence=0, sort_by_size=False):
        """
        Return ordered list of unused Item objects.
        """
        if not 0 <= min_confidence <= 100:
            raise ValueError("min_confidence must be between 0 and 100.")

        def by_name(item):
            return (str(item.filename).lower(), item.first_lineno)

        def by_size(item):
            return (item.size,) + by_name(item)

        unused_code = (
            self.unused_attrs
            + self.unused_classes
            + self.unused_funcs
            + self.unused_imports
            + self.unused_methods
            + self.unused_props
            + self.unused_vars
            + self.unreachable_code
        )

        confidently_unused = [
            obj for obj in unused_code if obj.confidence >= min_confidence
        ]

        return sorted(
            confidently_unused, key=by_size if sort_by_size else by_name
        )

    def report(
        self, min_confidence=0, sort_by_size=False, make_whitelist=False
    ):
        """
        Print ordered list of Item objects to stdout.
        """
        for item in self.get_unused_code(
            min_confidence=min_confidence, sort_by_size=sort_by_size
        ):
            print(
                item.get_whitelist_string()
                if make_whitelist
                else item.get_report(add_size=sort_by_size)
            )
            self.found_dead_code_or_error = True
        return self.found_dead_code_or_error

    @property
    def unused_classes(self):
        return _get_unused_items(self.defined_classes, self.used_names)

    @property
    def unused_funcs(self):
        return _get_unused_items(self.defined_funcs, self.used_names)

    @property
    def unused_imports(self):
        return _get_unused_items(self.defined_imports, self.used_names)

    @property
    def unused_methods(self):
        return _get_unused_items(self.defined_methods, self.used_names)

    @property
    def unused_props(self):
        return _get_unused_items(self.defined_props, self.used_names)

    @property
    def unused_vars(self):
        return _get_unused_items(self.defined_vars, self.used_names)

    @property
    def unused_attrs(self):
        return _get_unused_items(self.defined_attrs, self.used_names)

    def _log(self, *args):
        if self.verbose:
            print(*args)

    def _add_aliases(self, node):
        """
        We delegate to this method instead of using visit_alias() to have
        access to line numbers and to filter imports from __future__.
        """
        assert isinstance(node, (ast.Import, ast.ImportFrom))
        for name_and_alias in node.names:
            # Store only top-level module name ("os.path" -> "os").
            # We can't easily detect when "os.path" is used.
            name = name_and_alias.name.partition(".")[0]
            alias = name_and_alias.asname
            self._define(
                self.defined_imports,
                alias or name,
                node,
                confidence=90,
                ignore=_ignore_import,
            )
            if alias is not None:
                self.used_names.add(name_and_alias.name)

    def _handle_conditional_node(self, node, name):
        if utils.condition_is_always_false(node.test):
            self._define(
                self.unreachable_code,
                name,
                node,
                last_node=node.body
                if isinstance(node, ast.IfExp)
                else node.body[-1],
                message=f"unsatisfiable '{name}' condition",
                confidence=100,
            )
        elif utils.condition_is_always_true(node.test):
            else_body = node.orelse
            if name == "ternary":
                self._define(
                    self.unreachable_code,
                    name,
                    else_body,
                    message="unreachable 'else' expression",
                    confidence=100,
                )
            elif else_body:
                self._define(
                    self.unreachable_code,
                    "else",
                    else_body[0],
                    last_node=else_body[-1],
                    message="unreachable 'else' block",
                    confidence=100,
                )
            elif name == "if":
                # Redundant if-condition without else block.
                self._define(
                    self.unreachable_code,
                    name,
                    node,
                    message="redundant if-condition",
                    confidence=100,
                )

    def _define(
        self,
        collection,
        name,
        first_node,
        last_node=None,
        message="",
        confidence=DEFAULT_CONFIDENCE,
        ignore=None,
    ):
        def ignored(lineno):
            return (
                (ignore and ignore(self.filename, name))
                or _match(name, self.ignore_names)
                or noqa.ignore_line(self.noqa_lines, lineno, ERROR_CODES[typ])
            )

        last_node = last_node or first_node
        typ = collection.typ
        first_lineno = lines.get_first_line_number(first_node)

        if ignored(first_lineno):
            self._log(f'Ignoring {typ} "{name}"')
        else:
            last_lineno = lines.get_last_line_number(last_node)
            collection.append(
                Item(
                    name,
                    typ,
                    self.filename,
                    first_lineno,
                    last_lineno,
                    message=message,
                    confidence=confidence,
                )
            )

    def _define_variable(self, name, node, confidence=DEFAULT_CONFIDENCE):
        self._define(
            self.defined_vars,
            name,
            node,
            confidence=confidence,
            ignore=_ignore_variable,
        )

    def visit_arg(self, node):
        """Function argument"""
        self._define_variable(node.arg, node, confidence=100)

    def visit_AsyncFunctionDef(self, node):
        return self.visit_FunctionDef(node)

    def visit_Attribute(self, node):
        if isinstance(node.ctx, ast.Store):
            self._define(self.defined_attrs, node.attr, node)
        elif isinstance(node.ctx, ast.Load):
            self.used_names.add(node.attr)

    def visit_BinOp(self, node):
        """
        Parse variable names in old format strings:

        "%(my_var)s" % locals()
        """
        if (
            isinstance(node.left, ast.Str)
            and isinstance(node.op, ast.Mod)
            and self._is_locals_call(node.right)
        ):
            self.used_names |= set(re.findall(r"%\((\w+)\)", node.left.s))

    def visit_Call(self, node):
        # Count getattr/hasattr(x, "some_attr", ...) as usage of some_attr.
        if isinstance(node.func, ast.Name) and (
            (node.func.id == "getattr" and 2 <= len(node.args) <= 3)
            or (node.func.id == "hasattr" and len(node.args) == 2)
        ):
            attr_name_arg = node.args[1]
            if isinstance(attr_name_arg, ast.Str):
                self.used_names.add(attr_name_arg.s)

        # Parse variable names in new format strings:
        # "{my_var}".format(**locals())
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Str)
            and node.func.attr == "format"
            and any(
                kw.arg is None and self._is_locals_call(kw.value)
                for kw in node.keywords
            )
        ):
            self._handle_new_format_string(node.func.value.s)

    def _handle_new_format_string(self, s):
        def is_identifier(name):
            return bool(re.match(r"[a-zA-Z_][a-zA-Z0-9_]*", name))

        parser = string.Formatter()
        try:
            names = [name for _, name, _, _ in parser.parse(s) if name]
        except ValueError:
            # Invalid format string.
            names = []

        for field_name in names:
            # Remove brackets and their contents: "a[0][b].c[d].e" -> "a.c.e",
            # then split the resulting string: "a.b.c" -> ["a", "b", "c"]
            vars = re.sub(r"\[\w*\]", "", field_name).split(".")
            for var in vars:
                if is_identifier(var):
                    self.used_names.add(var)

    @staticmethod
    def _is_locals_call(node):
        """Return True if the node is `locals()`."""
        return (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "locals"
            and not node.args
            and not node.keywords
        )

    def visit_ClassDef(self, node):
        for decorator in node.decorator_list:
            if _match(
                utils.get_decorator_name(decorator), self.ignore_decorators
            ):
                self._log(
                    f'Ignoring class "{node.name}" (decorator whitelisted)'
                )
                break
        else:
            self._define(
                self.defined_classes, node.name, node, ignore=_ignore_class
            )

    def visit_FunctionDef(self, node):
        decorator_names = [
            utils.get_decorator_name(decorator)
            for decorator in node.decorator_list
        ]

        first_arg = node.args.args[0].arg if node.args.args else None

        if "@property" in decorator_names:
            typ = "property"
        elif (
            "@staticmethod" in decorator_names
            or "@classmethod" in decorator_names
            or first_arg == "self"
        ):
            typ = "method"
        else:
            typ = "function"

        if any(
            _match(name, self.ignore_decorators) for name in decorator_names
        ):
            self._log(f'Ignoring {typ} "{node.name}" (decorator whitelisted)')
        elif typ == "property":
            self._define(self.defined_props, node.name, node)
        elif typ == "method":
            self._define(
                self.defined_methods, node.name, node, ignore=_ignore_method
            )
        else:
            self._define(
                self.defined_funcs, node.name, node, ignore=_ignore_function
            )

    def visit_If(self, node):
        self._handle_conditional_node(node, "if")

    def visit_IfExp(self, node):
        self._handle_conditional_node(node, "ternary")

    def visit_Import(self, node):
        self._add_aliases(node)

    def visit_ImportFrom(self, node):
        if node.module != "__future__":
            self._add_aliases(node)

    def visit_Name(self, node):
        if (
            isinstance(node.ctx, (ast.Load, ast.Del))
            and node.id not in IGNORED_VARIABLE_NAMES
        ):
            self.used_names.add(node.id)
        elif isinstance(node.ctx, (ast.Param, ast.Store)):
            self._define_variable(node.id, node)

    def visit_Assign(self, node):
        if _assigns_special_variable__all__(node):
            assert isinstance(node.value, (ast.List, ast.Tuple))
            for elt in node.value.elts:
                if isinstance(elt, ast.Str):
                    self.used_names.add(elt.s)

    def visit_While(self, node):
        self._handle_conditional_node(node, "while")

    def visit_MatchClass(self, node):
        for kwd_attr in node.kwd_attrs:
            self.used_names.add(kwd_attr)

    def visit(self, node):
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, None)
        if self.verbose:
            lineno = getattr(node, "lineno", 1)
            line = self.code[lineno - 1] if self.code else ""
            self._log(lineno, ast.dump(node), line)
        if visitor:
            visitor(node)

        # There isn't a clean subset of node types that might have type
        # comments, so just check all of them.
        type_comment = getattr(node, "type_comment", None)
        if type_comment is not None:
            mode = (
                "func_type"
                if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
                else "eval"
            )
            self.visit(
                ast.parse(type_comment, filename="<type_comment>", mode=mode)
            )

        return self.generic_visit(node)

    def _handle_ast_list(self, ast_list):
        """
        Find unreachable nodes in the given sequence of ast nodes.
        """
        for index, node in enumerate(ast_list):
            if isinstance(
                node, (ast.Break, ast.Continue, ast.Raise, ast.Return)
            ):
                try:
                    first_unreachable_node = ast_list[index + 1]
                except IndexError:
                    continue
                class_name = node.__class__.__name__.lower()
                self._define(
                    self.unreachable_code,
                    class_name,
                    first_unreachable_node,
                    last_node=ast_list[-1],
                    message=f"unreachable code after '{class_name}'",
                    confidence=100,
                )
                return

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for _, value in ast.iter_fields(node):
            if isinstance(value, list):
                self._handle_ast_list(value)
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)


def main():
    config = make_config()
    vulture = Vulture(
        verbose=config["verbose"],
        ignore_names=config["ignore_names"],
        ignore_decorators=config["ignore_decorators"],
    )
    vulture.scavenge(config["paths"], exclude=config["exclude"])
    sys.exit(
        vulture.report(
            min_confidence=config["min_confidence"],
            sort_by_size=config["sort_by_size"],
            make_whitelist=config["make_whitelist"],
        )
    )



#! /usr/bin/env python

import pathlib
import re

import setuptools


def find_version(*parts):
    here = pathlib.Path(__file__).parent
    version_file = here.joinpath(*parts).read_text()
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]$", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("README.md") as f1, open("CHANGELOG.md") as f2:
    long_description = f1.read() + "\n\n" + f2.read()

setuptools.setup(
    name="vulture",
    version=find_version("vulture", "version.py"),
    description="Find dead code",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="dead-code-removal",
    author="Jendrik Seipp",
    author_email="jendrikseipp@gmail.com",
    url="https://github.com/jendrikseipp/vulture",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Quality Assurance",
    ],
    install_requires=["toml"],
    entry_points={"console_scripts": ["vulture = vulture.core:main"]},
    python_requires=">=3.6",
    packages=setuptools.find_packages(exclude=["tests"]),
    package_data={"vulture": ["whitelists/*.py"]},
)



from whitelist_utils import Whitelist

# NodeVisitor methods are called implicitly.
whitelist_node_visitor = Whitelist()

whitelist_node_visitor.visit_Assert
whitelist_node_visitor.visit_Assign
whitelist_node_visitor.visit_AsyncFor
whitelist_node_visitor.visit_AsyncFunctionDef
whitelist_node_visitor.visit_AsyncWith
whitelist_node_visitor.visit_Attribute
whitelist_node_visitor.visit_AugAssign
whitelist_node_visitor.visit_Await
whitelist_node_visitor.visit_BinOp
whitelist_node_visitor.visit_BoolOp
whitelist_node_visitor.visit_Bytes
whitelist_node_visitor.visit_Call
whitelist_node_visitor.visit_ClassDef
whitelist_node_visitor.visit_Compare
whitelist_node_visitor.visit_Constant
whitelist_node_visitor.visit_Delete
whitelist_node_visitor.visit_Dict
whitelist_node_visitor.visit_DictComp
whitelist_node_visitor.visit_ExceptHandler
whitelist_node_visitor.visit_Exec
whitelist_node_visitor.visit_Expr
whitelist_node_visitor.visit_Expression
whitelist_node_visitor.visit_ExtSlice
whitelist_node_visitor.visit_For
whitelist_node_visitor.visit_FunctionDef
whitelist_node_visitor.visit_GeneratorExp
whitelist_node_visitor.visit_Global
whitelist_node_visitor.visit_If
whitelist_node_visitor.visit_IfExp
whitelist_node_visitor.visit_Import
whitelist_node_visitor.visit_ImportFrom
whitelist_node_visitor.visit_Index
whitelist_node_visitor.visit_Interactive
whitelist_node_visitor.visit_Lambda
whitelist_node_visitor.visit_List
whitelist_node_visitor.visit_ListComp
whitelist_node_visitor.visit_MatchClass
whitelist_node_visitor.visit_Module
whitelist_node_visitor.visit_Name
whitelist_node_visitor.visit_NameConstant
whitelist_node_visitor.visit_Nonlocal
whitelist_node_visitor.visit_Num
whitelist_node_visitor.visit_Print
whitelist_node_visitor.visit_Raise
whitelist_node_visitor.visit_Repr
whitelist_node_visitor.visit_Return
whitelist_node_visitor.visit_Set
whitelist_node_visitor.visit_SetComp
whitelist_node_visitor.visit_Slice
whitelist_node_visitor.visit_Starred
whitelist_node_visitor.visit_Str
whitelist_node_visitor.visit_Subscript
whitelist_node_visitor.visit_Suite
whitelist_node_visitor.visit_Try
whitelist_node_visitor.visit_TryExcept
whitelist_node_visitor.visit_TryFinally
whitelist_node_visitor.visit_Tuple
whitelist_node_visitor.visit_UnaryOp
whitelist_node_visitor.visit_While
whitelist_node_visitor.visit_With
whitelist_node_visitor.visit_Yield
whitelist_node_visitor.visit_YieldFrom
whitelist_node_visitor.visit_alias
whitelist_node_visitor.visit_arg
whitelist_node_visitor.visit_arguments
whitelist_node_visitor.visit_comprehension
whitelist_node_visitor.visit_keyword



import string

string.Formatter.check_unused_args
string.Formatter.convert_field
string.Formatter.format
string.Formatter.format_field
string.Formatter.get_field
string.Formatter.get_value
string.Formatter.parse
string.Formatter.vformat
